// file      : odb/mysql/database.hxx
// author    : Boris Kolpackov <boris@codesynthesis.com>
// copyright : Copyright (c) 2009-2011 Code Synthesis Tools CC
// license   : GNU GPL v2; see accompanying LICENSE file

#ifndef ODB_MYSQL_DATABASE_HXX
#define ODB_MYSQL_DATABASE_HXX

#include <odb/pre.hxx>

#include <string>
#include <memory> // std::auto_ptr
#include <iosfwd> // std::ostream

#include <odb/database.hxx>

#include <odb/mysql/mysql.hxx>
#include <odb/mysql/version.hxx>
#include <odb/mysql/forward.hxx>
#include <odb/mysql/connection.hxx>
#include <odb/mysql/connection-factory.hxx>
#include <odb/mysql/transaction-impl.hxx>

#include <odb/details/shared-ptr.hxx>

#include <odb/mysql/details/export.hxx>

namespace odb
{
  namespace mysql
  {
    class LIBODB_MYSQL_EXPORT database: public odb::database
    {
    public:
      typedef mysql::connection connection_type;

    public:
      // In MySQL, NULL and empty string are treated as the same value
      // for all the arguments except passwd and socket.
      //
      database (const char* user,
                const char* passwd,
                const char* db,
                const char* host = 0,
                unsigned int port = 0,
                const char* socket = 0,
                unsigned long client_flags = 0,
                std::auto_ptr<connection_factory> =
                  std::auto_ptr<connection_factory> (0));

      database (const std::string& user,
                const std::string& passwd,
                const std::string& db,
                const std::string& host = "",
                unsigned int port = 0,
                const std::string* socket = 0,
                unsigned long client_flags = 0,
                std::auto_ptr<connection_factory> =
                  std::auto_ptr<connection_factory> (0));

      database (const std::string& user,
                const std::string* passwd,
                const std::string& db,
                const std::string& host = "",
                unsigned int port = 0,
                const std::string* socket = 0,
                unsigned long client_flags = 0,
                std::auto_ptr<connection_factory> =
                  std::auto_ptr<connection_factory> (0));

      database (const std::string& user,
                const std::string& passwd,
                const std::string& db,
                const std::string& host,
                unsigned int port,
                const std::string& socket,
                unsigned long client_flags = 0,
                std::auto_ptr<connection_factory> =
                  std::auto_ptr<connection_factory> (0));

      database (const std::string& user,
                const std::string* passwd,
                const std::string& db,
                const std::string& host,
                unsigned int port,
                const std::string& socket,
                unsigned long client_flags = 0,
                std::auto_ptr<connection_factory> =
                  std::auto_ptr<connection_factory> (0));

      // Extract the database parameters from the command line. The
      // following options are recognized:
      //
      // --user
      // --password
      // --database
      // --host
      // --port
      // --socket
      // --options-file
      //
      // For more information, see the output of the print_usage() function
      // below. If erase is true, the above options are removed from the
      // argv array and the argc count is updated accordingly. This
      // constructor may throw the cli_exception exception.
      //
      database (int& argc,
                char* argv[],
                bool erase = false,
                unsigned long client_flags = 0,
                std::auto_ptr<connection_factory> =
                  std::auto_ptr<connection_factory> (0));

      static void
      print_usage (std::ostream&);


    public:
      const char*
      user () const
      {
        return user_.c_str ();
      }

      const char*
      password () const
      {
        return passwd_;
      }

      const char*
      db () const
      {
        return db_.c_str ();
      }

      const char*
      host () const
      {
        return host_.c_str ();
      }

      unsigned int
      port () const
      {
        return port_;
      }

      const char*
      socket () const
      {
        return socket_;
      }

      unsigned long
      client_flags () const
      {
        return client_flags_;
      }

    public:
      using odb::database::execute;

      virtual unsigned long long
      execute (const char* statement, std::size_t length);

    public:
      virtual transaction_impl*
      begin ();

    public:
      details::shared_ptr<connection_type>
      connection ();

    public:
      virtual
      ~database ();

    private:
      std::string user_;
      std::string passwd_str_;
      const char* passwd_;
      std::string db_;
      std::string host_;
      unsigned int port_;
      std::string socket_str_;
      const char* socket_;
      unsigned long client_flags_;
      std::auto_ptr<connection_factory> factory_;
    };
  }
}

#include <odb/mysql/database.ixx>

#include <odb/post.hxx>

#endif // ODB_MYSQL_DATABASE_HXX
